/********************************************************/
/* Flash Linker Advance                                 */
/*   by Jeff Frohwein, 2001-Jun-28                      */
/* Compiled with DJGPP & linux                          */
/********************************************************/
// V1.0  - 01/06/28 - Original release
// V1.1  - 01/06/29 - Add -w option to slow down I/O transfer for some.
// V1.11 - 01/06/30 - Set ECP chipsets to proper parallel port mode.


// To compile source on linux:
//   cc -o fl fl.c -O2
// You must have root access to run this under linux.


// RAM Detect notes for dev. (Just ignore!)
//-----------------------------------------
// To detect backup type.
//  1. First check for eeprom.
//  2. Read byte from 0xe000000.
//  3. Write new byte to 0xe000000.
//  4. If diff, write back data & exit with SRAM detect flag.
//  5. If no diff get device Manuf ID for flash.
//  6. If no Manuf ID detected then report no cart backup available.

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef DJGPP
 #include <fcntl.h>
 #include <conio.h>
 #include <dos.h>
 #include <io.h>
 #define outp(p,v)  outportb(p,v); iodelay()
 #define inp(p)   inportb(p)
 #define getchr getch
#elif __linux__
// #include <sys/types.h>
// #include <unistd.h>
// #include <curses.h>
// #include <terms.h>
 //#include <termio.h>
 #include <termios.h>
// #include <sys/ioctl.h>
// #include <ncurses/term.h>
// #include <ncurses/curses.h>

 #include <signal.h>
 #include <asm/io.h>
 #define outp(p,v)  outb(v,p); iodelay()
 #define inp(p)   inb(p)
 #define getchr cnt_getch
 //ioctl_getch
#endif

//#define HEADER_LENGTH 0xc0
//#define OUTBUFLEN 256                   // Must be a multiple of 2! (ex:64,128,256...)

#define INTEL28F_BLOCKERASE 0x20
#define INTEL28F_CONFIRM    0xD0
#define INTEL28F_QUIRY      0x98
#define INTEL28F_READARRAY  0xff
#define INTEL28F_READSR     0x70
#define INTEL28F_RIC        0x90
#define INTEL28F_WRTOBUF    0xe8

#define SHARP28F_BLOCKERASE 0x20
#define SHARP28F_CONFIRM    0xD0
#define SHARP28F_WORDWRITE  0x10

#define u8      unsigned char
#define u16     unsigned int
#define u32     unsigned long
#define CONST_U8 const unsigned char


//int WaitDelay,WaitNCDelay;
unsigned SPPDataPort;
unsigned SPPStatPort;
unsigned SPPCtrlPort;
unsigned EPPAddrPort;
unsigned EPPDataPort;

int debug,verbose;
int EPPMode;
int DataSize16;
int Device;
int WaitDelay;

void iodelay (void)
   {
   int i;
   for (i=0; i<WaitDelay; i++)
      {
      i++;
      i--;
      }
   }

#ifdef __linux__
static int mbquit = 0;
static struct termios orig, new;
static int peek = -1;

void control_c(int n)
   {
   mbquit = 1;
   }

void InitLinuxKbhit (init)
   {
   if (init)
      {
      tcgetattr(0, &orig);
      new = orig;
      new.c_lflag &= ~ICANON;
      new.c_lflag &= ~ECHO;
      new.c_lflag &= ~ISIG;
      new.c_cc[VMIN] = 1;
      new.c_cc[VTIME] = 0;
      tcsetattr(0, TCSANOW, &new);
      }
   else
      {
      tcsetattr(0,TCSANOW, &orig);
      }
   }

int kbhit()
   {
   char ch;
   int nread;

   if (peek != -1) return 1;
   new.c_cc[VMIN]=0;
   tcsetattr(0, TCSANOW, &new);
   nread = read(0,&ch,1);
   new.c_cc[VMIN]=1;
   tcsetattr(0, TCSANOW, &new);

   if (nread == 1)
      {
      peek = ch;
      return 1;
      }

   return 0;
   }

int cnt_getch()
   {
   char ch;

   if (peek != -1)
      {
      ch = peek;
      peek = -1;
      return ch;
      }

   read(0,&ch,1);
   return ch;
   }

#endif

void ProgramExit (int code)
   {
#ifdef __linux__
   InitLinuxKbhit (0);
#endif
   exit(code);
   }

void usage(char *name)
   {
   char _small[255];
   char smaller[255];
   int i = 0;

   strcpy(_small, name);

#ifndef __linux__
   if (strchr(name, '.') != NULL)
      _small[strlen(_small)-4] = 0;               /* remove trailing file type */
#endif

   while ( (_small[strlen(_small)-i] != 0x2f) &&   /* loop until we find a / */
           ((strlen(_small)-i) > 0 ) )
      i++;

   if ((strlen(_small)-i) == 0) i++;

   strcpy (smaller, (char *)(&_small[strlen(_small)-i+1]) );

   fprintf (stderr, "GBA FLinker v1.11 by Jeff F.\n");
   fprintf (stderr, "Usage: %s [-c n][-d n][-h][-l n][-m][-p file][-s file][-w n]\n\n", smaller);

//   fprintf (stderr, "\t-b n\tBackup 32k bytes of SRAM into file\n");
   fprintf (stderr, "\t-c n\tSpecify chip size in mbits (8,16,32,64,128,256) (default=32)\n");
   fprintf (stderr, "\t-d n\tDump 256 bytes of ROM to screen (default: n=0)\n");
   fprintf (stderr, "\t-h\tThis help screen\n");
   fprintf (stderr, "\t-l n\tSpecify the parallel port to use (default is 1 = LPT1)\n");
   fprintf (stderr, "\t-m\tSet Standard Parallel Port (SPP) mode (default = EPP)\n");
   fprintf (stderr, "\t-p file\tProgram flash cart with file\n");
   fprintf (stderr, "\t-s file\tSave the cart into a file (Use -c to specify size)\n");
   fprintf (stderr, "\t-w n\tAdd delay to make transfer more reliable\n");
   }

void InitPort (int port)
   {
   int ECPRegECR;

   if (port == 1)
       SPPDataPort = 0x378;
   else
       SPPDataPort = 0x278;

   SPPStatPort = SPPDataPort + 1;
   SPPCtrlPort = SPPDataPort + 2;
   EPPAddrPort = SPPDataPort + 3;
   EPPDataPort = SPPDataPort + 4;
   ECPRegECR   = SPPDataPort + 0x402;

//#ifndef __linux__
   if (EPPMode)
      { outp (ECPRegECR, 4); }             // Set EPP mode for ECP chipsets
   else
      { outp (ECPRegECR, 0); }             // Set SPP mode for ECP chipsets
//#endif
   }

void l4020a4 (int reg)
   {
   outp (SPPCtrlPort, 1);
   outp (SPPDataPort, reg);
   outp (SPPCtrlPort, 9);
   outp (SPPCtrlPort, 1);
   }

void l4020dc (int i)
   {
   outp (SPPDataPort, i);
   outp (SPPCtrlPort, 3);
   outp (SPPCtrlPort, 1);
   }

void l402108 (int reg, int adr)
   {
   l4020a4 (reg);
   l4020dc (adr);
   }

int SPPReadByte (void)                      // 402124
   {
   int v;

   outp (SPPCtrlPort, 0);
   outp (SPPCtrlPort, 2);
   v = ((inp(SPPStatPort)) >> 3) & 0xf;
   outp (SPPCtrlPort, 6);
   v += (((inp(SPPStatPort)) << 1) & 0xf0);
   outp (SPPCtrlPort, 0);

   return (v);
   }

void l402188 (int reg)
   {
   outp (SPPCtrlPort, 1);
   outp (EPPAddrPort, reg);
   }

void l4021a8 (int reg, int adr)
   {
   l402188 (reg);
   outp (SPPCtrlPort, 1);
   outp (EPPDataPort, adr);
   }

void l4021d0 (int i)
   {
   outp (SPPCtrlPort, 1);

   if (EPPMode)
      l402188 (i);
   else
      l4020a4 (i);
   }

void l402200 (int reg, int adr)
   {
   if (EPPMode)
      l4021a8 (reg, adr);
   else
      l402108 (reg, adr);
   }

void l402234 (void)
   {
   if (EPPMode)
      l402200 (4, 0x83);
   else
      l402200 (4, 0xc3);
   }

void l40226c (void)
   {
   if (EPPMode)
      l402200 (4, 7);
   else
      l402200 (4, 0x47);
   }

void l4022d0 (int i)
   {
   if (EPPMode)
      {
      outp (EPPDataPort, i);
      if (DataSize16)
         {
         outp (EPPDataPort, (i>>8));
         }
      }
   else
      {
      l4020dc(i);
      l4020dc(i>>8);
      }
   }

int PPReadByte (void)                  // 40234c
   {
   int v;

   if (EPPMode)
      v = inp (EPPDataPort);
   else
      v = SPPReadByte ();

   return (v);
   }

int ReadFlash (void)                      // 402368
   {
   int v = 0;

   if (EPPMode)
      {
      v = inp (EPPDataPort);
      if (DataSize16)
         v += (inp (EPPDataPort) << 8);
      }
   else
      {
      v = SPPReadByte (); //402124
      v += (SPPReadByte () << 8);
      }
   return (v);
   }

void SetCartAddr (int addr)             // 4023cc
   {
   l402200 (2,addr>>16);
   l402200 (1,addr>>8);
   l402200 (0,addr);
   }

void WriteFlash (int addr, int data)    // 402414
   {
   SetCartAddr (addr);
   l4021d0 (3);
   l4022d0 (data);
   }

void WriteRepeat (int addr, int data, int count)
   {
   int i;
   for (i=0; i<count; i++)
      WriteFlash (addr, data);
   }


void VisolyModePreamble (void)  // 402438
   {
   l40226c ();
   WriteRepeat (0x987654, 0x5354, 1);
   WriteRepeat ( 0x12345, 0x1234, 500);
   WriteRepeat (  0x7654, 0x5354, 1);
   WriteRepeat ( 0x12345, 0x5354, 1);
   WriteRepeat ( 0x12345, 0x5678, 500);
   WriteRepeat (0x987654, 0x5354, 1);
   WriteRepeat ( 0x12345, 0x5354, 1);
   WriteRepeat (0x765400, 0x5678, 1);
   WriteRepeat ( 0x13450, 0x1234, 1);
   WriteRepeat ( 0x12345, 0xabcd, 500);
   WriteRepeat (0x987654, 0x5354, 1);
   }

void SetVisolyFlashRWMode (void)
   {
   VisolyModePreamble ();
   WriteFlash (0xf12345, 0x9413);
   }

void SetVisolyBackupRWMode (int i)                     // 402550
   {
   VisolyModePreamble ();
   WriteFlash (0xa12345, i>>1);
   }

void l402684 (void)
   {
   outp (SPPStatPort, 1);
   l40226c ();
   }

void LinkerInit (void)
   {
   l402684 ();
//   l402200 (2,0x12);
//   l402200 (1,0x34);
//   l402200 (0,0x56);
   l4021d0 (2);
   outp (SPPCtrlPort, 0);
   outp (SPPCtrlPort, 4);
   }

//void l4027c4 (void)
//   {
//   LinkerInit();
//   }

int ReadStatusRegister (void)                     // 402dd8
   {
   int v;
   WriteFlash (0,INTEL28F_READSR);
   outp (SPPCtrlPort, 0);
   v = ReadFlash () & 0xff;
   return (v);
   }

void DumpSRAM (void)                    // 4046f4
   {
   int i,j,k,v;

   i=1;
   SetVisolyBackupRWMode (i);
   l402234 ();

   for (j=0; j<0x80; j++)
      {
      l402200 (0, 0);
      if (i==1)
         l402200 (1, j);
      else
         l402200 (1, j|0x80);
      l402200 (2, 0);
      l4021d0 (3);
      outp (SPPCtrlPort, 0);

      for (k=0; k<0x100; k++)
         {
         v = PPReadByte ();
         printf("%x ", v);
         }
      }
   printf("\n");
   }

void BackupROM (FILE *fp, int SizekW)
   {
   u16 valw;
   u32 i;

   WriteFlash (0, INTEL28F_READARRAY);         // Set flash (intel 28F640J3A) Read Mode

   SetCartAddr (0);                            // Set cart base addr to 0
   l4021d0 (3);

   outp (SPPCtrlPort, 0);

   for (i = 0; i < (SizekW); i++)
      {
      valw = ReadFlash ();
      fputc(valw & 0xff, fp);
      fputc(valw >> 8, fp);
      }
   }

void dump (u8 BaseAdr)
   {
//   unsigned char low, high;
   int i;
   u8 First = 1;
   u16 v;
   u8 val1,val2;
   u8 Display[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

   WriteFlash (0, INTEL28F_READARRAY);         // Set flash (intel 28F640J3A) Read Mode

   SetCartAddr (BaseAdr<<7);                   // Set cart base addr to read
   l4021d0 (3);

   outp (SPPCtrlPort, 0);

   for(i=0; i<128; i++)
      {
      if (First == 1)
         {
         if (i*2 < 256) printf("0");
         if (i*2 < 16) printf("0");
         printf("%hx - ",(i*2));
         First = 0;
         }

      v = ReadFlash();
      val2 = v >> 8;
      val1 = v & 255;

      if ((val1 > 31) & (val1 < 127))
       Display[(i & 7)*2] = val1;
      else
       Display[(i & 7)*2] = 46;

      if (val1 < 16)
         printf("0");
      printf("%hx ",val1);

      if ((val2 > 31) & (val2 < 127))

       Display[(i & 7)*2+1] = val2;
      else
       Display[(i & 7)*2+1] = 46;

      if (val2 < 16)
         printf("0");
      printf("%hx ",val2);

      if ((i & 7)==7)
         {
         First = 1;
         printf("   %3s",(&Display[0]));
         printf("\n");
         }
      }
   }

void CheckForFC (void)                  // 402f40
   {
   int Manuf;

   LinkerInit ();
   SetVisolyFlashRWMode ();
   WriteFlash (0, INTEL28F_RIC);          // Read Identifier codes from flash.
                                          // Works for intel 28F640J3A & Sharp LH28F320BJE.

   SetCartAddr (0);                     // Set cart addr to 0
   l4021d0 (3);

   outp (SPPCtrlPort, 0);

   Manuf = ReadFlash ();
   Device = ReadFlash ();

   if (Manuf == 0x2e)
      printf("Manufacturer ID = Standard ROM\n");
   else
      printf("Manufacturer ID = 0x%x: Device ID = 0x%x: ", Manuf, Device);

   switch (Device)
      {
      case 0x17:
         printf("(intel i28F640J3A)\n");
         break;
      case 0x18:
         printf("(intel i28F128J3A)\n");
         break;
      case 0xe2:
         printf("(Sharp LH28F320BJE)\n");
         break;
      case 0xea00:
         // Regular ROM
         Device = 0;
         break;
      default:
         printf("(Unknown)\n");
         Device = 0;
      }
   }

void ProgramIntelFlash (FILE *fp)
   {
   int i,j,k;
   int addr = 0;
   int done;
   int FileSize = -1;

   // Get file size
   while (!feof(fp))
      {
      (void) fgetc(fp);
      FileSize++;
      }
   rewind(fp);

   printf("Erasing flash cart...");
#ifndef __linux__
   fflush(stdout);       // Force printf to print to screen even though no \n
#endif

   j=1;
   for (i=0; i<(FileSize>>1); i=i+65536)
      {
      WriteFlash (addr, INTEL28F_BLOCKERASE);          // Erase a block
      WriteFlash (addr, INTEL28F_CONFIRM);             // Comfirm block erase
      while ((ReadStatusRegister () & 0x80)==0)
         {
         }
      }
//   ProgramExit(1);
   printf("\nProgramming Visoly flash cart...");
#ifndef __linux__
   fflush(stdout);       // Force printf to print to screen even though no \n
#endif

   while (!feof(fp))
      {
      done = 0;
      while (!done)
         {
         WriteFlash (addr, INTEL28F_WRTOBUF);
         outp (SPPCtrlPort, 0);

         done = ReadFlash() & 0x80;
         }

      WriteFlash (addr, 0xf);              // Write 0xf+1 words

      SetCartAddr (addr);                           // Set cart base addr to 0
      l4021d0 (3);

      j = 0;
      for (i=0; i<16; i++)
         {
         if (j != EOF)
            j = fgetc(fp);
         k = j;
         if (j != EOF)
            j = fgetc(fp);
         k += (j << 8);
         l4022d0 (k);

         }
      addr += 16;
      WriteFlash (0, INTEL28F_CONFIRM);             // Comfirm block write
      }

   WriteFlash (0, INTEL28F_READARRAY);
   outp (SPPCtrlPort, 0);

   printf("\nDone.\n");
   }

void ProgramSharpFlash (FILE *fp)
   {
   int i,j;
   int k = 0;
   int addr = 0;
//   int done;
   int FileSize = -1;

   // Get file size
   while (!feof(fp))
      {
      (void) fgetc(fp);
      FileSize++;
      }
   rewind(fp);

   printf("Erasing flash cart...");
#ifndef __linux__
   fflush(stdout);       // Force printf to print to screen even though no \n
#endif
//   j = 0;
   for (i=0; i<(FileSize>>1); i=i+32768)
      {
//      printf("%d,%ld\n", ++j,i);
      WriteFlash (addr, SHARP28F_BLOCKERASE);          // Erase a block
      WriteFlash (addr, SHARP28F_CONFIRM);             // Comfirm block erase
      while ((ReadStatusRegister () & 0x80)==0)
         {
//         printf("RSR = %d\n", ReadStatusRegister ());
         }
//      printf(" %d,%ld\n", j,i);
      }

   printf("\nProgramming Nintendo flash cart...");
#ifndef __linux__
   fflush(stdout);       // Force printf to print to screen even though no \n
#endif

   while (!feof(fp))
      {
      j = fgetc(fp);
      if (j != EOF)
         {
         k = fgetc(fp);
         }
      i = ((k & 0xff)<<8)+(j & 0xff);

      while ((ReadStatusRegister () & 0x80)==0)
         {
         }

      WriteFlash(addr, SHARP28F_WORDWRITE);
      WriteFlash(addr, i);
      addr += 1;
      }

   WriteFlash (0, INTEL28F_READARRAY);
   outp (SPPCtrlPort, 0);

   printf("\nDone.\n");
   }

void VerifyFlash (FILE *fp)
   {
   int addr = 0;
   int CompareFail = 0;
   int k = 0;
   int i,j,m,n;

   WriteFlash (0, INTEL28F_READARRAY);         // Set flash (intel 28F640J3A) Read Mode

   SetCartAddr (0);                            // Set cart base addr to read
   l4021d0 (3);

   outp (SPPCtrlPort, 0);

   j=0;
   while (!feof(fp))
      {
      j = fgetc(fp);
      if (j != EOF)
         {
         k = fgetc(fp);

         i = ReadFlash();
         m = i & 0xff;
         n = i >> 8;

         if (m != j)
            {
            printf("Address %x - Cartridge %x : File %hx\n", addr, m, j);
            CompareFail = 1;
            }
         if (n != k)
            {
            printf("Address %x - Cartridge %x : File %hx\n", addr+1, n, k);
            CompareFail = 1;
            }
         addr += 2;
         }
      }
   if (CompareFail == 0)
      printf("%d bytes compared ok.\n", addr);
   }

#ifdef __linux__
void *old = NULL;

void clear_sig(void)
   {
   if (old)
      signal(SIGINT, old);
   }
#endif

int main(int argc, char **argv)
{
    int arg,i;
    u8 Base = 0;
    FILE *fp;
    char fname[20],fname2[20];
    int OptB = 0;
    int OptD = 0;
    int OptP = 0;
    int OptS = 0;
    int OptV = 0;
    int port = 1;
    int ChipSize = 32;

#ifdef __linux__
   InitLinuxKbhit (1);
#endif

    if(argc < 2) {
        usage(argv[0]);
        ProgramExit(1);
	}

    debug = 0;
    verbose = 1;
    EPPMode = 1;
    DataSize16 = 1;
    WaitDelay = 0;

    for (arg = 1; arg < argc; arg++) {
		if(argv[arg][0] != '-') {
            usage(argv[0]);
            ProgramExit(1);
		}
		switch(argv[arg][1]) {
//            case 'b':
//                    OptB=1;
//                    break;
            case 'c':
                    // Set cart size
                    ChipSize = (u16)(atoi(argv[++arg]));
                    if ( (ChipSize != 8) &&
                         (ChipSize != 16) &&
                         (ChipSize != 32) &&
                         (ChipSize != 64) &&
                         (ChipSize != 128) &&
                         (ChipSize != 256) )
                       {
                       fprintf(stderr, "ERROR: Chip size must be 8,16,32,64,128 or 256.\n");
                       ProgramExit(1);
                       }
                    break;
            case 'd':
                    // Dump 256 bytes to screen
                    if (argv[++arg] != NULL)
                       Base = (u8)(atoi(argv[arg]));
                    printf("Base address : %hx\n",Base*256);
                    OptD = 1;
                    break;
            case 'm':
                    // Set SPP mode
                    EPPMode = 0;
                    break;
            case 's':
                    // Backup flash cart
                    strcpy(fname, argv[++arg]);
                    OptS = 1;
                    break;
            case 'p':
                    // Program flash cart
                    strcpy(fname, argv[++arg]);
                    OptP = 1;
                    break;
            case 'v':
                    // Verify flash cart
                    strcpy(fname2, argv[++arg]);
                    OptV = 1;
                    break;
            case 'l':
                    i = atoi(argv[++arg]);
                    if ((i==1) || (i==2))
                       port = i;
                    else
                       {
                       fprintf(stderr, "ERROR: Only LPT1 & LPT2 are supported");
                       ProgramExit(1);
                       }
			    	break;
            case 'w':
                    WaitDelay = atoi(argv[++arg]);
                    break;
            default:
                    usage(argv[0]);
                    ProgramExit(1);
		}
	}

#ifdef __linux__
    atexit(clear_sig);
    old = signal(SIGINT, control_c);
    if (getuid() && geteuid()) {
        printf("ERROR: No privileges to use I/O. Obtain root access & retry\n");
        exit(1);
    }
    ioperm((port == 1) ? 0x378 : 0x278, 5, 1);
    iopl (3);  // Get access to ECP ECR reg = SPPDataPort + 0x402
    if (getgid() && !getegid())
        setegid(getgid());

    if (getuid() && !geteuid())
        seteuid(getuid());
#endif

   InitPort(port);

   CheckForFC ();

   if (OptB) DumpSRAM ();

   if (OptD) dump (Base);

   if ( (OptP) &&
        (Device == 0) )
      {
      fprintf(stderr, "ERROR: Device type unrecognized\n");
      ProgramExit(1);
      }

   if (OptP)
      {
#ifndef __linux__
      if (strchr(fname, '.') == NULL)
         strcat(fname, ".gba");
#endif
      if ((fp = fopen(fname, "rb")) == NULL)
         {
         fprintf(stderr, "ERROR trying to open file '%s'\n", fname);
         ProgramExit(1);
         }
      printf("Programming flash with file '%s'.\n", fname);

      if (Device == 0xe2)
         ProgramSharpFlash (fp);
      else
         ProgramIntelFlash (fp);
      fclose(fp);
      }

   if (OptS)
      {
#ifndef __linux__
      if (strchr(fname, '.') == NULL)
         strcat(fname, ".gba");
#endif
      if ((fp = fopen(fname, "wb")) == NULL)
         {
         fprintf(stderr, "ERROR trying to open file '%s'\n", fname);
         ProgramExit(1);
         }
      printf("Backing up %d mbits of ROM to file '%s'. Please wait...", ChipSize, fname);
#ifndef __linux__
      fflush(stdout);       // Force printf to print to screen even though no \n
#endif

      BackupROM (fp, ChipSize<<16);
      printf("\n");
      fclose(fp);
      }

   if (OptV)
      {
#ifndef __linux__
      if (strchr(fname2, '.') == NULL)
         strcat(fname2, ".gba");
#endif
      if ((fp = fopen(fname2, "rb")) == NULL)
         {
         fprintf(stderr, "ERROR trying to open file '%s'\n", fname2);
         ProgramExit(1);
         }
      printf("Comparing flash with file '%s'.\n", fname2);

      VerifyFlash (fp);
      fclose(fp);
      }

   ProgramExit(0);
   exit(0);
}